/* -*- mode: c++; c-basic-offset: 4; -*- */
#include "lsmp_driver.hh"
#include "iSMbuf.hh"
#include "loc_list.hh"
#include "chan_list.hh"
#include <stdexcept>

#include "framecpp/IFrameStream.hh"
#include "framecpp/FrTOC.hh"

#include <fstream>
#include <sstream>
#include <iostream>

using namespace std;
using namespace sends;

typedef FrameCPP::Common::FrameBuffer< iSMbuf > frame_buffer;

//======================================  Constants
//
//  The lsmp_timeout variable is a constant value that specify the timeout 
//  period for all lsmp accesses.
//
static const double lsmp_timeout(20.0);


//======================================  Constructor
lsmp_driver::lsmp_driver(void) 
    : mLSMPbuffer(0), mActiveChan(0), mNextGPS(0), mState(0)
{
    mFrameBuf = 0;
}

//======================================  Constructor
lsmp_driver::lsmp_driver(const std::string& partition) 
    : mPartition(partition), mLSMPbuffer(0), mActiveChan(0), mNextGPS(0),
      mState(0)
{
    mFrameBuf = 0;
}

//======================================  Destructor
lsmp_driver::~lsmp_driver(void) {
    stop_consumer();
}

//======================================  Get channel index
//
//   Return the largest entry with channel index less than or equal to
//   the specified index.
index_type 
lsmp_driver::chanlist_index(chan_index c) const {
    index_type lo  = 0;
    index_type hi  = mChanList.size();
    while (lo < hi) {
	index_type mid = (lo + hi) >> 1;
	if (mChanList[mid]._chan >= c) hi = mid;
	else if (lo != mid)            lo = mid;
	else                           break;
    }
    return hi;
}

//======================================  Dump the channel list entry
void 
lsmp_driver::dump_chanlist(std::ostream& out) const {
    index_type N = mChanList.size();
    out << "Current " << name() << " reserved channel list, size = " 
	<< N << endl;
    out << "index channel       use-count" << endl;
    for (index_type i=0; i<N; ++i) {
	chan_index c = mChanList[i]._chan;
	out << i << "  " << Channel_List[c].name() << " (" << c << ") " 
	    << mChanList[i]._count << endl;
    }
}

//======================================  Find named channel entry
lsmp_driver::chan_node& 
lsmp_driver::find_channel(chan_index c) {
    index_type i = chanlist_index(c);
    if (i<mChanList.size() && mChanList[i]._chan == c) return mChanList[i];
    dump_chanlist(cerr);
    string msg = string("Online channel: ") + Channel_List[c].name() 
	       + " not in reserved list";
    throw runtime_error(msg);
}

//======================================  Find named channel entry
const lsmp_driver::chan_node& 
lsmp_driver::find_channel(chan_index c) const {
    index_type i = chanlist_index(c);
    if (i<mChanList.size() && mChanList[i]._chan == c) return mChanList[i];
    dump_chanlist(cerr);
    string msg = string("Online channel: ") + Channel_List[c].name()
	       + " not in reserved list";
    throw runtime_error(msg);
}

//======================================  Add named channel to channel list
//
//  This function looks to see if a channel is in the list, If not, an 
//  appropriate channel entry is added in the correct position with a 
//  zero use count.
//
lsmp_driver::chan_node& 
lsmp_driver::insert_channel(chan_index c) {
    //----------------------------------  Get a pointer, see if the channel
    index_type i = chanlist_index(c);
    index_type N = mChanList.size();
    if (i == N || mChanList[i]._chan != c) {
	chan_node last(c);
	bool push_last = true;
	for (index_type j=i; j<N; ++j) {
	    if (mChanList[j]._count) {
		chan_node t = last;
		last = mChanList[j];
		mChanList[j] = t;
	    } else {
		mChanList[j] = last;
		push_last = false;
		break;
	    }
	}
	if (push_last) mChanList.push_back(last);
    }
    return mChanList[i];
}

//======================================  Get the current online gps time
string 
lsmp_driver::name(void) const {
    return string("lsmp:") + mPartition;
}

//======================================  Pick total time from TOC
inline gps_type
frameDt(const FrameCPP::Common::FrTOC& toc) {
    int N = toc.nFrame();
    double dt = 0;
    for (int i=0; i<N; ++i) dt += toc.dt()[i];
    return gps_type(dt);
}

//======================================  Get the current online gps time
gps_type
lsmp_driver::online_gps(void) {
    if (!mNextGPS) {
	thread::semlock lock(mGate);

	//------------------------------  Just in case something happened
	if (mNextGPS) return mNextGPS;
	if (!mLSMPbuffer && !start_consumer()) {
	    throw runtime_error(string("Cannot start consumer: ")+name());
	}

	//------------------------------  Clear timeout if data have arrived
	if (test_state(kStateTimeOut) && mLSMPbuffer->waitBuf(true)) {
	    set_state(kStateTimeOut, false);
	}

	//------------------------------  Read until we find new data or error
	while (!mNextGPS && ready()) {
	    if (!mLSMPbuffer->timedWait(lsmp_timeout)) {
		switch (errno) {
		case EINTR:
		    break;
		case EAGAIN:
		    set_state(kStateTimeOut);
		    break;
		default:
		    set_state(kStateError);
		    break;
		}
	    } else if (mLSMPbuffer->latest()) {
		mNextGPS = mLSMPbuffer->eventid();
	    } else {
		if (debug()) cerr << "skipping gps=" 
				  << mLSMPbuffer->eventid()
				  << endl;
		mLSMPbuffer->relse();
	    }
	}
    }
    //if (!mNextGPS) {
    //	throw runtime_error(string("Cannot get start time for ") + part());
    //}
    return mNextGPS;
}

//======================================  Read data into segments
int
lsmp_driver::read_data(const loc_list& ll, gps_type start, 
		       gps_type stop, const channel_vect& cv) 
{
    // Is the floc stop time really meaningful?
    // if (stop > ll.back().stop()) stop = ll.back().stop();

    //----------------------------------  Loop over frames
    int rc = 0;
    while (mNextGPS < stop) {

	//------------------------------  Take control if I'm the first.
	thread::semlock lock(mGate);
	if (mNextGPS >= stop) continue;


	channel_vect v;
	segment_vect segv;
	try {

	    //----------------------  Find the start of the data.
	    if (!mLSMPbuffer) {
		throw runtime_error("Shared memory partition not open");
	    }
	    if (test_state(kStateTimeOut) && mLSMPbuffer->waitBuf(true)) {
		set_state(kStateTimeOut, false);
	    }

	    //----------------------  Scan data for start time.
	    gps_type buffer_gps  = mLSMPbuffer->eventid();
	    while (ready() && buffer_gps < mNextGPS) {
		mLSMPbuffer->timedWait(lsmp_timeout);
		buffer_gps = mLSMPbuffer->eventid();
		if (!buffer_gps) break;
		if (buffer_gps < mNextGPS) mLSMPbuffer->relse();
	    }
	    if (!buffer_gps || buffer_gps != mNextGPS) {
		//ostringstream msg;
		cerr << "Can't find start time (" << mNextGPS 
		    << ") in partition " << name() << endl;
		//throw runtime_error(msg.str());
		rc = -1;
		break;
	    }

	    //----------------------  build a channel list.
	    thread::semlock chan_lock(mChanMutex, thread::semlock::rdlock);
	    v.clear();
	    v.reserve(mActiveChan);
	    index_type nChan = mChanList.size();
	    for (index_type i=0; i<nChan; i++) {
		if (mChanList[i]._count) v.push_back(mChanList[i]._chan);
	    }

	    //----------------------  reserve segments for this frame
	    gps_type dt = ll.back().duration();
	    gps_type end_frame = mNextGPS + dt;
	    if (get_segments(mNextGPS, end_frame, v, segv)) {
		//------------------  Open the frame
		if (debug() >= 2) cerr << "Read data from gps: " 
				       << mNextGPS << endl;
		FrameCPP::IFrameStream frdr(mFrameBuf);	    
		frdr.SetAutoDeleteBuffer( false );

		//------------------  Check frame duration
		gps_type dt_frame = frameDt(*frdr.GetTOC());
		if (dt_frame != dt) {
		    cerr << "Frame length in " << name() << " (" << dt_frame 
			 << ") is not expected (" << dt << ")" << endl;
		    if (dt_frame < dt) {
			end_frame = mNextGPS + dt_frame;
		    } else {
			cerr << "Reject frame because of length error" << endl;
			rc = -1;
			break;
		    }
		}

		//------------------  Read a frame
		if (read_frame(frdr, segv)) rc = -1;
	    }
	    mNextGPS = end_frame;
	    mLSMPbuffer->relse();

	//------------------------------  On error, release segments and 
	//                                rethrow exception
	} catch (...) {
	    release_segments(segv);
	    throw;
	}
	release_segments(segv);
    }
    return rc;
}

//======================================  Release the specified channel
int 
lsmp_driver::release_channel(chan_index c) {
    thread::semlock(mChanMutex, thread::semlock::wrlock);
    if (--find_channel(c)._count == 0) mActiveChan--;
    if (debug() >= 2) cerr << "lsmp_driver: channel " << Channel_List[c].name() 
			   << " released, active=" << mActiveChan << endl;
    if (!mActiveChan) {
	mNextGPS = 0;
	stop_consumer();
    }
    return 0;
}

//======================================  Request the specified channel
int 
lsmp_driver::request_channel(chan_index c) {
    thread::semlock(mChanMutex, thread::semlock::wrlock);
    if (insert_channel(c)._count++ == 0) mActiveChan++;
    if (debug() >= 2) cerr << "lsmp_driver: channel " << Channel_List[c].name() 
			   << " requested, active=" << mActiveChan << endl;
    return 0;
}

//======================================  Start an lsmp consumer
void
lsmp_driver::set_state(state_bits bit, bool val) {
    if (val) mState |=  (1 << bit);
    else     mState &= ~(1 << bit);
}

//======================================  Start an lsmp consumer
bool 
lsmp_driver::start_consumer(void) {
    if (!mLSMPbuffer) {
	frame_buffer* buf = new frame_buffer(ios::in);
	mLSMPbuffer  =  dynamic_cast<iSMbuf*>(buf);
	mFrameBuf = buf;
	if (mLSMPbuffer == NULL) {
	    set_state(kStateOffline, true);
	} else if (mLSMPbuffer->open(part()) == NULL) {
	    set_state(kStateOffline, true);
	    delete mFrameBuf;
	    mFrameBuf   = NULL;
	    mLSMPbuffer = NULL;
	} else {
	    mState = 0;
	}
    }
    return ready();
}

//======================================  Stop/disconnect an lsmp consumer.
bool 
lsmp_driver::stop_consumer(void) {
    if (mLSMPbuffer) {
	mLSMPbuffer->relse();
	delete mFrameBuf;
	mFrameBuf = 0;
	mLSMPbuffer = 0;
    }
    mNextGPS = 0;
    return true;
}

//======================================  Test whether frame exists
bool 
lsmp_driver::test_available(const loc_list& fl, gps_type t) const 
{
    return mLSMPbuffer != 0;
}

#include "framecpp/Common/IOStream.icc"
 template class FrameCPP::Common::FrameBuffer< iSMbuf >;
